[Openvpn-devel,v2,3/9] Add connection_established as state in tls_multi->context_auth

Message ID 20210520151148.2565578-3-arne@rfc2549.org
State Accepted
Delegated to: Antonio Quartulli
Headers show
Series [Openvpn-devel,v2,1/9] Move auth_token_state from multi to key_state | expand

Commit Message

Arne Schwabe May 20, 2021, 5:11 a.m. UTC
the socket_info->connection_establish is set through
link_socket_set_outgoing_addr when we reach FULL_SYNC. This patch
introduces a new state in context_auth that replaces the
connection_established state for TLS connections. This make the state
machine easier to understand.

Patch v2: fix p2p mode server without (without ncp)

Signed-off-by: Arne Schwabe <arne@rfc2549.org>
---
 src/openvpn/forward.c    |  6 +++---
 src/openvpn/forward.h    | 13 ++++++++++++-
 src/openvpn/multi.c      | 15 ++++++++-------
 src/openvpn/occ.c        |  2 +-
 src/openvpn/openvpn.h    |  4 +++-
 src/openvpn/push.c       |  2 +-
 src/openvpn/ssl.c        | 38 ++++++++++++++++++++++++++++----------
 src/openvpn/ssl_common.h | 12 +++++++++---
 8 files changed, 65 insertions(+), 27 deletions(-)

Comments

Antonio Quartulli June 10, 2021, 10:32 p.m. UTC | #1
Hi,

On 20/05/2021 17:11, Arne Schwabe wrote:
> the socket_info->connection_establish is set through
> link_socket_set_outgoing_addr when we reach FULL_SYNC. This patch
> introduces a new state in context_auth that replaces the
> connection_established state for TLS connections. This make the state
> machine easier to understand.
> 
> Patch v2: fix p2p mode server without (without ncp)
> 
> Signed-off-by: Arne Schwabe <arne@rfc2549.org>

Compile zoo is happy.

Performed various tests, including:
* classic server/client
* p2p+tls with ncp
* p2p+tls without ncp

All passed and renegotiation also worked as expected.

The patch looks sane (already reviewed v1 and the added change makes
sense) and it's a reasonable step forward towards making the state
machine simpler to understand.


Acked-by: Antonio Quartulli <antonio@openvpn.net>
Gert Doering June 24, 2021, 3:35 a.m. UTC | #2
Stared at the patch, seems to make sense.

Tested on the t_server test rig, and behaves nicely - both alone, and
together with 4/9 (the final patch needed to fix the CVE).  Test rig has
p2mp and p2p servers.

Client tests also succeded (unsurprisingly).

As discussed on IRC, added refererences to CVE-2020-15078 to the commit 
message.  Also added some comments about the CAS_* changes to the 
commit message, which were missing.

Your patch has been applied to the master branch.

commit d01277608a248f31df3fde1883eba6dd8d16a1e4
Author: Arne Schwabe
Date:   Thu May 20 17:11:42 2021 +0200

     Add connection_established as state in tls_multi->context_auth

     Signed-off-by: Arne Schwabe <arne@rfc2549.org>
     Acked-by: Antonio Quartulli <antonio@openvpn.net>
     Message-Id: <20210520151148.2565578-3-arne@rfc2549.org>
     URL: https://www.mail-archive.com/openvpn-devel@lists.sourceforge.net/msg22419.html
     Signed-off-by: Gert Doering <gert@greenie.muc.de>


--
kind regards,

Gert Doering

Patch

diff --git a/src/openvpn/forward.c b/src/openvpn/forward.c
index e302d8e0a..692ca7271 100644
--- a/src/openvpn/forward.c
+++ b/src/openvpn/forward.c
@@ -280,7 +280,7 @@  void
 check_connection_established(struct context *c)
 {
 
-    if (CONNECTION_ESTABLISHED(c))
+    if (connection_established(c))
     {
         /* if --pull was specified, send a push request to server */
         if (c->c2.tls_multi && c->options.pull)
@@ -536,7 +536,7 @@  encrypt_sign(struct context *c, bool comp_frag)
      * has not yet succeeded. In non-TLS tls_multi mode is not defined
      * and we always pass packets.
      */
-    if (c->c2.tls_multi && c->c2.tls_multi->multi_state != CAS_SUCCEEDED)
+    if (c->c2.tls_multi && c->c2.tls_multi->multi_state < CAS_CONNECT_DONE)
     {
         c->c2.buf.len = 0;
     }
@@ -971,7 +971,7 @@  process_incoming_link_part1(struct context *c, struct link_socket_info *lsi, boo
          * has not yet succeeded. In non-TLS mode tls_multi is not defined
          * and we always pass packets.
          */
-        if (c->c2.tls_multi && c->c2.tls_multi->multi_state != CAS_SUCCEEDED)
+        if (c->c2.tls_multi && c->c2.tls_multi->multi_state < CAS_CONNECT_DONE)
         {
             c->c2.buf.len = 0;
         }
diff --git a/src/openvpn/forward.h b/src/openvpn/forward.h
index 2a67c1445..3461e6422 100644
--- a/src/openvpn/forward.h
+++ b/src/openvpn/forward.h
@@ -445,6 +445,17 @@  io_wait(struct context *c, const unsigned int flags)
     }
 }
 
-#define CONNECTION_ESTABLISHED(c) (get_link_socket_info(c)->connection_established)
+static inline bool
+connection_established(struct context *c)
+{
+    if (c->c2.tls_multi)
+    {
+        return c->c2.tls_multi->multi_state >= CAS_CONNECT_DONE;
+    }
+    else
+    {
+        return get_link_socket_info(c)->connection_established;
+    }
+}
 
 #endif /* FORWARD_H */
diff --git a/src/openvpn/multi.c b/src/openvpn/multi.c
index def5dd40b..3f9710134 100644
--- a/src/openvpn/multi.c
+++ b/src/openvpn/multi.c
@@ -674,7 +674,8 @@  multi_close_instance(struct multi_context *m,
 #ifdef ENABLE_MANAGEMENT
     set_cc_config(mi, NULL);
 #endif
-    if (mi->context.c2.tls_multi->multi_state == CAS_SUCCEEDED)
+
+    if (mi->context.c2.tls_multi->multi_state >= CAS_CONNECT_DONE)
     {
         multi_client_disconnect_script(mi);
     }
@@ -775,7 +776,7 @@  multi_create_instance(struct multi_context *m, const struct mroute_addr *real)
         goto err;
     }
 
-    mi->context.c2.tls_multi->multi_state = CAS_PENDING;
+    mi->context.c2.tls_multi->multi_state = CAS_NOT_CONNECTED;
 
     if (hash_n_elements(m->hash) >= m->max_clients)
     {
@@ -2407,7 +2408,7 @@  multi_client_connect_late_setup(struct multi_context *m,
     mi->reporting_addr_ipv6 = mi->context.c2.push_ifconfig_ipv6_local;
 
     /* set context-level authentication flag */
-    mi->context.c2.tls_multi->multi_state = CAS_SUCCEEDED;
+    mi->context.c2.tls_multi->multi_state = CAS_CONNECT_DONE;
 
     /* authentication complete, calculate dynamic client specific options */
     if (!multi_client_set_protocol_options(&mi->context))
@@ -2649,9 +2650,9 @@  multi_connection_established(struct multi_context *m, struct multi_instance *mi)
 
             case CC_RET_DEFERRED:
                 /*
-                 * we already set client_connect_status to DEFERRED_RESULT or
+                 * we already set multi_status to DEFERRED_RESULT or
                  * DEFERRED_NO_RESULT. We just return
-                 * from the function as having client_connect_status
+                 * from the function as having multi_status
                  */
                 return;
 
@@ -3003,8 +3004,8 @@  multi_process_post(struct multi_context *m, struct multi_instance *mi, const uns
         {
             /* connection is "established" when SSL/TLS key negotiation succeeds
              * and (if specified) auth user/pass succeeds */
-            if (is_cas_pending(mi->context.c2.tls_multi->multi_state)
-                && CONNECTION_ESTABLISHED(&mi->context))
+
+            if (is_cas_pending(mi->context.c2.tls_multi->multi_state))
             {
                 multi_connection_established(m, mi);
             }
diff --git a/src/openvpn/occ.c b/src/openvpn/occ.c
index 3ff351aa0..c169838e1 100644
--- a/src/openvpn/occ.c
+++ b/src/openvpn/occ.c
@@ -185,7 +185,7 @@  check_send_occ_req_dowork(struct context *c)
 void
 check_send_occ_load_test_dowork(struct context *c)
 {
-    if (CONNECTION_ESTABLISHED(c))
+    if (connection_established(c))
     {
         const struct mtu_load_test *entry;
 
diff --git a/src/openvpn/openvpn.h b/src/openvpn/openvpn.h
index 6e4651cf5..2b74975fa 100644
--- a/src/openvpn/openvpn.h
+++ b/src/openvpn/openvpn.h
@@ -206,7 +206,7 @@  struct context_1
 
 
 static inline bool
-is_cas_pending(enum client_connect_status cas)
+is_cas_pending(enum multi_status cas)
 {
     return cas == CAS_PENDING || cas == CAS_PENDING_DEFERRED
            || cas == CAS_PENDING_DEFERRED_PARTIAL;
@@ -237,6 +237,8 @@  struct context_2
 
     struct link_socket *link_socket;     /* socket used for TCP/UDP connection to remote */
     bool link_socket_owned;
+
+    /** This variable is used instead link_socket->info for P2MP UDP childs */
     struct link_socket_info *link_socket_info;
     const struct link_socket *accept_from; /* possibly do accept() on a parent link_socket */
 
diff --git a/src/openvpn/push.c b/src/openvpn/push.c
index 85f9488a5..2d621e472 100644
--- a/src/openvpn/push.c
+++ b/src/openvpn/push.c
@@ -867,7 +867,7 @@  process_incoming_push_request(struct context *c)
         send_auth_failed(c, client_reason);
         ret = PUSH_MSG_AUTH_FAILURE;
     }
-    else if (c->c2.tls_multi->multi_state == CAS_SUCCEEDED)
+    else if (c->c2.tls_multi->multi_state >= CAS_CONNECT_DONE)
     {
         time_t now;
 
diff --git a/src/openvpn/ssl_common.h b/src/openvpn/ssl_common.h
index 53325e556..8a65ab984 100644
--- a/src/openvpn/ssl_common.h
+++ b/src/openvpn/ssl_common.h
@@ -504,12 +504,18 @@  struct tls_session
 /* client authentication state, CAS_SUCCEEDED must be 0 since
  * non multi code path still checks this variable but does not initialise it
  * so the code depends on zero initialisation */
-enum client_connect_status {
-    CAS_SUCCEEDED=0,
+
+/* CAS_NOT_CONNECTED is the initial state for every context. When the *first*
+ * tls_session reaches S_ACTIVE, this state machine moves to CAS_PENDING (server)
+ * or CAS_CONNECT_DONE (client/p2p) as clients skip the stages associated with
+ * connect scripts/plugins */
+enum multi_status {
+    CAS_NOT_CONNECTED,
     CAS_PENDING,
     CAS_PENDING_DEFERRED,
     CAS_PENDING_DEFERRED_PARTIAL,   /**< at least handler succeeded, no result yet*/
     CAS_FAILED,
+    CAS_CONNECT_DONE,
 };
 
 
@@ -548,7 +554,7 @@  struct tls_multi
 
     int n_sessions;             /**< Number of sessions negotiated thus
                                  *   far. */
-    enum client_connect_status multi_state;
+    enum multi_status multi_state;
 
     /*
      * Number of errors.
diff --git a/src/openvpn/ssl.c b/src/openvpn/ssl.c
index 7eb356ddf..9f3f83f16 100644
--- a/src/openvpn/ssl.c
+++ b/src/openvpn/ssl.c
@@ -2381,19 +2381,22 @@  key_method_2_write(struct buffer *buf, struct tls_multi *multi, struct tls_sessi
      * If we're a p2mp server to allow NCP, the first key
      * generation is postponed until after the connect script finished and the
      * NCP options can be processed. Since that always happens at after connect
      * script options are available the CAS_SUCCEEDED status is identical to
      * NCP options are processed and we have no extra state for NCP finished.
      */
-    if (session->opt->server && (session->opt->mode != MODE_SERVER
-            || multi->multi_state == CAS_SUCCEEDED))
+    if (ks->authenticated > KS_AUTH_FALSE && session->opt->server
+        && ((session->opt->mode == MODE_SERVER && multi->multi_state >= CAS_CONNECT_DONE)
+            || (session->opt->mode == MODE_POINT_TO_POINT && !session->opt->pull)))
     {
-        if (ks->authenticated > KS_AUTH_FALSE)
+        /* if key_id >= 1, is a renegotiation, so we use the already established
+         * parameters and do not need to delay anything. */
+
+        /* key-id == 0 and multi_state >= CAS_CONNECT_DONE is a special case of
+         * the server reusing the session of a reconnecting client. */
+        if (!tls_session_generate_data_channel_keys(session))
         {
-            if (!tls_session_generate_data_channel_keys(session))
-            {
-                msg(D_TLS_ERRORS, "TLS Error: server generate_key_expansion failed");
-                goto error;
-            }
+            msg(D_TLS_ERRORS, "TLS Error: server generate_key_expansion failed");
+            goto error;
         }
     }
 
@@ -2801,6 +2804,21 @@  tls_process(struct tls_multi *multi,
                 /* Set outgoing address for data channel packets */
                 link_socket_set_outgoing_addr(to_link_socket_info, &ks->remote_addr, session->common_name, session->opt->es);
 
+                /* Check if we need to advance the tls_multi state machine */
+                if (multi->multi_state == CAS_NOT_CONNECTED)
+                {
+                    if (session->opt->mode == MODE_SERVER)
+                    {
+                        /* On a server we continue with running connect scripts next */
+                        multi->multi_state = CAS_PENDING;
+                    }
+                    else
+                    {
+                        /* Skip the connect script related states */
+                        multi->multi_state = CAS_CONNECT_DONE;
+                    }
+                }
+
                 /* Flush any payload packets that were buffered before our state transitioned to S_ACTIVE */
                 flush_payload_buffer(ks);
 
@@ -2381,19 +2381,22 @@  key_method_2_write(struct buffer *buf, struct tls_multi *multi, struct tls_sessi
      * generation is postponed until after the connect script finished and the
      * NCP options can be processed. Since that always happens at after connect
-     * script options are available the CAS_SUCCEEDED status is identical to
-     * NCP options are processed and we have no extra state for NCP finished.
+     * script options are available the CAS_CONNECT_DONE status is identical to
+     * NCP options are processed and do not wait for NCP being finished.
      */